//
// 自動迷路作成のコーディングにおいて、@zurazurataicho様の「C言語で穴掘り法を使った自動迷路作成を実装してみた」を流用しております。
// https://qiita.com/zurazurataicho/items/1435b7236b6b1ca34334
//

#include <time.h>

#include "misc.h"

#include "map.h"
#include "gameType.h"
#include "time.h"
#include "nextStage.h"

#define SPACE 0x20

typedef enum {
	eWALL,	// 壁
	eROAD,	// 道
	eCHAR,	// 操作キャラ
	eGOAL,	// ゴール
	eSTAR,	// ★
	eCLOCK,
} CHIP;

typedef struct {
	int x;
	int y;
} POINT;

typedef struct {
	int stage;
	int starNum;
} PARAM;

static const PARAM kParam[] = {
	{ 1, 3 },
	{ 2, 4 },
	{ 3, 5 },
};

// mapの幅、高さ。奇数にすること
#define W (35)
#define H (21)

// マップ
static CHIP map[H][W];

// 操作キャラのx,y座標
static int s_charX;
static int s_charY;

// ゴールのx,y座標
static int s_goalX;
static int s_goalY;

static int starLength;

// 4方向に合わせて進むx,yの値

static const POINT kDir[] = {
	{ 0, -1 },	// 上
	{ 1, 0 },	// 右
	{ 0, 1 },	// 下
	{ -1, 0 }	// 左
};

// ★を管理する構造体
static POINT s_star[5];

// ◆を管理する構造体
typedef struct {
	int x;
	int y;
} CLOCK;
static CLOCK s_clock[2];

// カーソルを行頭に移す
static void topLineCursor() {
	printf("\x1b[1G");
}

// カーソルを画面先頭に戻す
static void topScreenCursor() {
	printf("\x1b[1;1H");
}

// 奇数の乱数を作る
static int randOdd(int mod) {
	int r = 1 + (rand() % mod);
	if (r % 2 == 0) {
		r++;
	}
	if (mod < r) {
		r = mod - 1;
	}
	return r;
}

// 迷路を作る
static void makeMaze(int x, int y) {
	const int DIR_MAX = 4;
	int dirStart = rand() % DIR_MAX;
	int d = dirStart;

	while (1) {
		int nx = x + (kDir[d].x * 2);
		int ny = y + (kDir[d].y * 2);
		if (nx < 0 || W <= nx
		 || ny < 0 || H <= ny
		 || map[ny][nx] != eWALL) {
			d++;
			d %= DIR_MAX;
			if (d == dirStart) {
				return;
			}
			continue;
		}
		map[y + kDir[d].y][x + kDir[d].x] = eROAD;
		map[ny][nx] = eROAD;
		makeMaze(nx, ny);
		dirStart = rand() % DIR_MAX;
		d = dirStart;
	}
}

// 四隅のうち一番遠い所をゴールにする
static void findGoal(int x, int y, int *pGx, int *pGy) {
	typedef struct {
		int x;
		int y;
	} GOAL;
	const GOAL kGoal[] = {
		{ 1, 1 },						// 左上
		{ (W - 1) - 1, 1 },				// 右上
		{ (W - 1) - 1, (H - 1) - 1 },	// 右下
		{ 1, (H - 1) - 1 }				// 左下
	};
	int i, maxIdx;
	int distance, maxDistance;

	maxIdx = 0;
	maxDistance = 0;
	for (i = 0; i < ARRAYOF(kGoal); i++) {
		distance = ((x - kGoal[i].x) * (x - kGoal[i].x))
			+ ((y - kGoal[i].y) * (y - kGoal[i].y));
		if (maxDistance < distance) {
			maxDistance = distance;
			maxIdx = i;
		}
	}
	*pGx = kGoal[maxIdx].x;
	*pGy = kGoal[maxIdx].y;
}

// ★を配置する
static void setupStar() {
	int i;
	int x, y;

	for (i = 0; i < ARRAYOF(kParam); i++) {
		if (kParam[i].stage == getStage()) {
			starLength = kParam[i].starNum;
		}
	}

	for (i = 0; i < starLength; i++) {
		do {
			x = rand() % W;
			y = rand() % H;
		} while (map[y][x] != eROAD);
		s_star[i].x = x;
		s_star[i].y = y;
		map[s_star[i].y][s_star[i].x] = eSTAR;
	}
}

// ◆を配置する
void setupClock() {
	int i;
	int x, y;

	for (i = 0; i < ARRAYOF(s_clock); i++) {
		do {
			x = rand() % W;
			y = rand() % H;
		} while (map[y][x] != eROAD);
		s_clock[i].x = x;
		s_clock[i].y = y;
		map[s_clock[i].y][s_clock[i].x] = eCLOCK;
	}
}

// マップを初期化する
void setupMap() {
	int y, x;

	// eWALLで初期化する
	for (y = 0; y < H; y++) {
		for (x = 0; x < W; x++) {
			map[y][x] = eWALL;
		}
	}

	// 迷路を作る
	x = randOdd(W - 2);
	y = randOdd(H - 2);
	makeMaze(x, y);

	// スタートを決める
	s_charX = x;
	s_charY = y;
	map[s_charY][s_charX] = eCHAR;

	// ゴールを決める
	findGoal(x, y, &s_goalX, &s_goalY);
	map[s_goalY][s_goalX] = eGOAL;

	// ★を配置する
	setupStar();
	// ◆を配置する
	setupClock();
}

// 表示効果をデフォルトに戻す
static void clearEffect() {
	printf("\x1b[0m");
}

// 壁を表示する
static void displayWall() {
	printf("■");
    myCursorRight();  // カーソルを1文字分進める
}
// 道を表示する
static void displayRoad() {
//	printf("　");
//  myCursorRight();  // カーソルを1文字分進める
    printf("  ");   // 全角スペース1個分を半角スペース2個分に変える
}
// 操作キャラを表示する
static void displayChar() {
	printf("\x1b[32m");	// 緑
	printf("●");
    myCursorRight();  // カーソルを1文字分進める
}
// ゴールを表示する
static void displayGoal() {
	printf("\x1b[34m");	// 青
	printf("■");
    myCursorRight();  // カーソルを1文字分進める
}
// ★を表示する
static void displayStar() {
	printf("\x1b[33m");	// 黄
	printf("★");
    myCursorRight();  // カーソルを1文字分進める
}
// ◆を表示する
static void displayClock() {
	printf("\x1b[31m");	// 赤
	printf("◆");
    myCursorRight();  // カーソルを1文字分進める
}

// チップを表示する
static void displayChip(CHIP c) {
	typedef struct {
		CHIP c;
		void(*pDisplay)();
	} DISPLAY;
	const DISPLAY kDisplay[] = {
		{ eWALL, displayWall },	// 壁
		{ eROAD, displayRoad },	// 道
		{ eCHAR, displayChar },	// 操作キャラ
		{ eGOAL, displayGoal },	// ゴール
		{ eSTAR, displayStar },	// ★
		{ eCLOCK, displayClock }, // ◆
	};
	int i;

	for (i = 0; i < ARRAYOF(kDisplay); i++) {
		if (kDisplay[i].c == c) {
			(*kDisplay[i].pDisplay)();
			clearEffect();
			return;
		}
	}
}

// マップ画面を表示する
static void displayMap() {
	int y, x;

	// mapを描く
	for (y = 0; y < H; y++) {
		for (x = 0; x < W; x++) {
			displayChip(map[y][x]);
		}
		printf("\n");
	}
}

// 残りのスターの数を返す
static int remainingStar() {
	int i;
	int num = 0;

	for (i = 0; i < starLength; i++) {
		if (map[s_star[i].y][s_star[i].x] == eSTAR) {
			num++;
		}
	}

	return num;
}

// 残りのスターの数を表示する
void displayRemainingStar() {
	printf("    残りスター数 : %d", remainingStar());
}

// 今のステージ数を表示する
void displayStage() {
	printf("    ステージ : %d", getStage());
}

//クリアかどうかチェックする
static int clearCheck() {
	if (s_goalX == s_charX && s_goalY == s_charY) {
		if (remainingStar() == 0) {
			// クリアした
			return 1;
		}
	}
	return 0;
}

// キャラ移動できるなら移動する
static void moveChar(int keyCode) {
	typedef struct {
		int keyCode;
		const POINT *pDir;
	} MOVE;
	const MOVE kMove[] = {
		{ 'w', &kDir[0] },	// 上
		{ 'd', &kDir[1] },	// 右
		{ 's', &kDir[2] },	// 下
		{ 'a', &kDir[3] }	// 左
	};
	int i;
	int x, y;

	for (i = 0; i < ARRAYOF(kMove); i++) {
		if (keyCode == kMove[i].keyCode) {
			break;
		}
	}
	if (i == ARRAYOF(kMove)) {
		return;
	}

	x = s_charX + kMove[i].pDir->x;
	y = s_charY + kMove[i].pDir->y;
	if (map[y][x] == eWALL) {
		return;	// 移動しない
	}

	if (map[y][x] == eCLOCK) {
		extendedTime();
	}

	// ★を道に置き換えて移動先を★に変える
	map[s_charY][s_charX] = eROAD;
	s_charX = x;
	s_charY = y;
	map[s_charY][s_charX] = eCHAR;

	//キャラとゴールが重なっていないのにゴールがないとき
	if (!(s_goalX == s_charX && s_goalY == s_charY)) {
		if (map[s_goalY][s_goalX] != eGOAL) {
			map[s_goalY][s_goalX] = eGOAL;
		}
	}
	return;
}

// ゲーム画面を操作する
GmType gameDisplay() {
	char ch;

	displayMap();

	while (1) {

		// ガイド表示を更新する
		countSec();

		topLineCursor();
		displaySec();
		displayRemainingStar();
		displayStage();
        myCursorClear();    // カーソルから後ろを消す

		if (getCountSec() <= 0) {
			printf("\n");
			return GM_TIMEOVER;
		}

		if (mykbhit()) {
			ch = mygetch();
			if (ch == SPACE) {
				return GM_MENU;
			}

			// キャラ移動できるなら移動する
			moveChar(ch);

			topScreenCursor();

			displayMap();

		}
		// ゲームクリアかどうかチェックする
		if (clearCheck() == 1) {
			if (getStage() == 3) {
				return GM_CLEAR;
			}

			return GM_STAGECLEAR;
		}
	}
}
